<?php

namespace App\Console\Commands;

use Doctrine\ORM\EntityNotFoundException;
use Illuminate\Console\Command;
use TranquilCity\TranquilPavement\Application\Services\TranquilPlaceService;
use TranquilCity\TranquilPavement\Domain\Models\TranquilPlace;
use TranquilCity\TranquilPavement\Domain\Models\User;
use TranquilCity\TranquilPavement\Domain\Repositories\TranquilPlaceRepository;
use TranquilCity\TranquilPavement\Domain\Repositories\UserRepository;
use TranquilCity\TranquilPavement\Infrastructure\Services\TranquilFacebook;
use TranquilCity\TranquilPavement\Infrastructure\Services\Twitter;

class ImportTranquilPlacesCommand extends Command
{
    /**
     * --update updates the urls for all images, which change periodically
     *
     * @var string
     */
    protected $signature = 'tp:import:places {--update}';

    /**
     * The console command description.
     *
     * @var string
     */
    protected $description = 'Import tranquil places from Instagram and Twitter';

    private $updateAll;
    private $users;
    private $placeRepository;
    private $tranquilPlaceService;
    private $twitter;
    private $facebook;
    private $instagramAccessToken;

    public function __construct(UserRepository $users,
                                TranquilPlaceRepository $placeRepository,
                                TranquilPlaceService $tranquilPlaceService,
                                TranquilFacebook $facebook,
                                Twitter $twitter)
    {
        parent::__construct();

        $this->users = $users;
        $this->placeRepository = $placeRepository;
        $this->tranquilPlaceService = $tranquilPlaceService;
        $this->facebook = $facebook;
        $this->twitter = $twitter;
        $this->instagramAccessToken = env('INSTAGRAM_API_TOKEN');
    }

    /**
     * Execute the console command.
     *
     * @return void
     */
    public function handle()
    {
        $this->updateAll = $this->input->getOption('update');
        $this->output->writeln('Import started ' . date('d/m/Y H:i:s', time()));
        /** @var TranquilPlace[] $oldTwitterPlaces */
        $oldTwitterPlaces = $this->placeRepository->findAllTwitter();
        $twitterHiddenMap = [];
        foreach ($oldTwitterPlaces as $place) {
            $twitterHiddenMap[$place->getSlug()] = $place->isHidden();
        }
        try {
            $this->setTwitterAccessToken();
            $instagramPlaces = $this->importFromInstagram();
            if (!$this->updateAll) {
                $twitterPlaces = $this->importFromTwitter();
            }
            if ($this->updateAll) {
                $this->refreshInstagramImages();
                $this->refreshTwitterImages();
            }
            if ($this->updateAll) {
                $this->output->writeln('Reordering places (this may take some time)');
                $this->tranquilPlaceService->reorder();
            } else {
                $this->tranquilPlaceService->order(array_merge($instagramPlaces, $twitterPlaces));
            }
        } catch (\Exception $e) {
            $this->output->writeln('Error: ' . $e->getMessage());
        }
        /** @var TranquilPlace[] $newTwitterPlaces */
        $newTwitterPlaces = $this->placeRepository->findAllTwitter();
        foreach ($newTwitterPlaces as $place) {
            $slug = $place->getSlug();
            if (!array_key_exists($slug, $twitterHiddenMap)) {
                continue;
            }
            $wasHidden = $twitterHiddenMap[$slug];
            $currentHidden = $place->isHidden();
            if ($wasHidden) {
                $this->output->writeln('Already hidden ' . $slug);
            } else if ($currentHidden) {
                $this->output->writeln('Hid ' . $slug);
            } else {
                $this->output->writeln('Did not hide ' . $slug);
            }
        }
    }

    /**
     * @throws EntityNotFoundException
     */
    private function importFromInstagram()
    {
        $places = [];
        $media = $this->facebook->getTagMedia('tranquilcitylondon');
        foreach ($media as $m) {
            $place = $this->handleMedia((object) $m);
            if ($place) {
                $places[] = $place;
            }
        }
        return $places;
    }

    private function handleMedia($datum)
    {
        /** @var TranquilPlace $place */
        $place = $this->placeRepository->findByThirdPartyId($datum->slug);
        if ($place) {
            $this->output->writeln('Already loaded ' . $datum->slug);
            if (!$place || !$place->hasGeolocation()) {
                $this->tranquilPlaceService->geocode($place);
                return $place;
            }
            return false;
        } else {
            $place = $this->tranquilPlaceService->createFromInstagramData($datum);
            $this->output->writeln('Added ' . $datum->slug);
            return $place;
        }
    }

    private function refreshInstagramImages ()
    {
        /** @var TranquilPlace[] $places */
        $places = $this->placeRepository->findAllInstagram();
        foreach ($places as $place) {
            $slug = $place->getSlug();
            if (strpos($slug, 'p/') === 0) {
                $slug = substr($slug, 2);
                $place->setSlug($slug);
            }
            try {
                $imageUrl = $this->fetchImageUrl($slug);
                if ($imageUrl && strlen($imageUrl) > 20) {
                    $place->setImageUrl($imageUrl);
                    $place->setHidden(false);
                    $this->output->writeLn('Refreshed ' . $slug);
                }
            } catch (\Exception $e) {
                $this->output->writeln('Could not refresh ' . $slug);
                if (strpos($e->getMessage(), '404') !== false) {
                    $this->output->writeln('Hiding ' . $slug);
                    $place->setHidden(true);
                }
            }
            $this->placeRepository->save($place);
        }
    }

    /**
     * @throws EntityNotFoundException
     */
    private function refreshTwitterImages ()
    {
        $this->setTwitterAccessToken();
        /** @var TranquilPlace[] $places */
        $places = $this->placeRepository->findAllTwitter();
        foreach ($places as $place) {
            $id = $place->getTwitterId();
            try {
                $this->handleStatus($id);
            } catch (\Exception $e) {
                $this->output->writeln('Could not refresh ' . $id . ': ' . $e->getMessage());
                if ($e->getCode() === 144) { // Tweet does not exist
                    $this->output->writeln('Hiding ' . $id);
                    $place->setHidden(true);
                }
            }
            $this->placeRepository->save($place);
        }
    }

    private function fetchImageUrl ($slug)
    {
        $url = "https://graph.facebook.com/v8.0/instagram_oembed?url=https://www.instagram.com/p/$slug/&access_token=" . $this->instagramAccessToken;

        $response = file_get_contents($url);

        $data = json_decode($response, true);

        return empty($data['thumbnail_url']) ? null : $data['thumbnail_url'];
    }

    /**
     * @throws \Exception
     */
    private function importFromTwitter()
    {
        $user = $this->setTwitterAccessToken();
        $complete = false;
        $limit = 10;
        $maxId = null;
        $places = [];
        while (!$complete) {
            $media = $this->twitter->getTagMediaPaginated('tranquilcitylondon', $limit, $maxId);
            if (property_exists($media, 'statuses') && $media->statuses) {
                foreach ($media->statuses as $status) {
                    try {
                        $added = $this->handleStatus($status->id_str);
                        $complete = !$added && !$this->updateAll;
                        if ($added) {
                            $places[] = $added;
                        }
                    } catch (\Exception $e) {
                        $this->output->writeln('Error handling ' . $status->id . ': ' . $e->getMessage());
                    }
                    if (!$maxId || $status->id < $maxId) {
                        $maxId = $status->id;
                    }
                }
                // decrease the maxId by one as it is inclusive, so otherwise the last tweet would be fetched twice
                // 64-bit PHP 7 can do this no problem
                $maxId--;
            } else {
                if (property_exists($media, 'errors') && $media->errors) {
                    $this->output->writeln('Failed to fetch from Twitter. Removing token from database.');
                    $error = $media->errors[0];
                    if ($error && $error->code === 89) {
                        $user->setTwitterToken(null);
                        $user->setTwitterTokenSecret(null);
                        $this->users->save($user);
                    }
                }
                $complete = true;
            }
            if (!$complete) {
                $this->output->writeln('Sleeping for 10 seconds');
                sleep(10);
            }
        }
        return $places;
    }

    /**
     * @throws EntityNotFoundException
     */
    private function setTwitterAccessToken()
    {
        /** @var User $user */
        $user = $this->users->getWithTwitterToken();
        if (!$user) {
            throw new EntityNotFoundException('Could not find a twitter token in the database');
        }

        $twitterAccessToken = $user->getTwitterToken();
        $twitterSecretToken = $user->getTwitterTokenSecret();
        if (!$twitterAccessToken || !$twitterSecretToken) {
            throw new EntityNotFoundException('Could not find a twitter token in the database');
        }

        $this->twitter->setAccessToken($twitterAccessToken, $twitterSecretToken);
        return $user;
    }

    /**
     * @param $id_str
     * @return mixed
     * @throws \Exception
     */
    private function handleStatus($id_str)
    {
        sleep(1); // avoid rate limiting
        $status = $this->twitter->getTweetById($id_str);
        if (property_exists($status, 'errors')) {
            throw new \Exception($status->errors[0]->message, $status->errors[0]->code);
        }
        $twitterId = $status->id_str;
        /** @var TranquilPlace $place */
        $place = $this->placeRepository->findByThirdPartyId($twitterId);
        $update = $this->updateAll;
        if ($place) {
            $this->output->writeln('Already loaded ' . $twitterId);
            if (!$update) {
                return false;
            }
            $this->output->writeln('Updating ' . $twitterId);
            $this->tranquilPlaceService->updateFromTwitterData($place, $status);
            return false;
        } else {
            $place = $this->tranquilPlaceService->createFromTwitterData($twitterId, $status);
            $this->output->writeln('Added ' . $twitterId);
            return $place;
        }
    }
}
