<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use Doctrine\DBAL\Exception\UniqueConstraintViolationException;
use Illuminate\Http\Request;
use Illuminate\Http\Response;
use League\Csv\Writer;
use TranquilCity\TranquilPavement\Application\Services\Geocoder;
use TranquilCity\TranquilPavement\Application\Services\TranquilPlaceService;
use TranquilCity\TranquilPavement\Domain\Models\Point;
use TranquilCity\TranquilPavement\Domain\Models\TranquilPlace;
use TranquilCity\TranquilPavement\Domain\Repositories\TranquilPlaceRepository;
use TranquilCity\TranquilPavement\Infrastructure\Services\Instagram;
use TranquilCity\TranquilPavement\Infrastructure\Services\Twitter;

class TranquilPlacesController extends Controller
{
    public function get(Request $request, TranquilPlaceRepository $repository)
    {
        /** @var TranquilPlace $place */
        $place = $repository->findById($request->route('uid'));

        if (!$place) {
            return Response::create('Not Found', 404);
        }

        return $place->toArray();
    }

    public function list(Request $request, TranquilPlaceRepository $repository)
    {
        $order = $request->get('order');
        if ($order !== 'createdTime') {
            $order = 'defaultIndex';
        }
        $places = $repository->findAll($order);

        $placesAsArrays = array_map(
            function (TranquilPlace $place) {
                return $place->toArray();
            },
            $places
        );

        return $placesAsArrays;
    }

    /**
     * Download the Tranquil Places as a CSV
     *
     * @param TranquilPlaceRepository $repository
     * @return Response
     * @throws \League\Csv\CannotInsertRecord
     * @throws \TypeError
     */
    public function download(TranquilPlaceRepository $repository)
    {
        $places = $repository->findAll();

        $placesAsArrays = array_map(
            function (TranquilPlace $place) {
                return $place->toCSVRow();
            },
            $places
        );

        $header = array_keys((new TranquilPlace())->toCSVRow());

        $csv = Writer::createFromPath("php://temp", "w");
        $csv->setOutputBOM(Writer::BOM_UTF8);
        $csv->insertOne($header);
        $csv->insertAll($placesAsArrays);

        return Response::create($csv->getContent(), 200, [
            'Content-type' => 'text/csv',
            'Content-Disposition' => 'attachment; filename=tranquil-places.csv'
        ]);
    }

    public function search(Request $request, TranquilPlaceRepository $repository, Geocoder $geocoder)
    {
        $location = $request->input('q');
        if ($location) {
            $searchLocation = $geocoder->geocode($location);
        } else {
            $latitude = (float)$request->input('latitude');
            $longitude = (float)$request->input('longitude');
            if ($latitude && $longitude) {
                $searchLocation = new Point($latitude, $longitude);
            } else {
                $searchLocation = new Point(51.5029927, -0.2046901);
            }
        }

        if (!$searchLocation) {
            return Response::create('We could not find your location.', 404);
        }

        if (
            $searchLocation->getLatitude() < 51.2 ||
            $searchLocation->getLatitude() > 51.8 ||
            $searchLocation->getLongitude() > 0.4 ||
            $searchLocation->getLongitude() < -0.65
        ) {
            return Response::create('Please choose a London postcode.', 404);
        }

        $placesWithDistances = $repository->findByLocation($searchLocation, 'defaultIndex');

        $placesAsArrays = array_map(
            function ($placeWithDistance) {
                $place = $placeWithDistance['place']->toArray();
                $place['distance'] = $placeWithDistance['distance'];
                return $place;
            },
            $placesWithDistances
        );

        return [
            'geolocation' => $searchLocation->toArray(),
            'places' => $placesAsArrays
        ];
    }

    public function create(Request $request, Instagram $instagram, Twitter $twitter, TranquilPlaceService $tranquilPlaceService)
    {
        $url = $request->get('url');
        if (!$url) {
            return Response::create('Invalid Request', 400);
        }

        $isInstagram = strpos($url, 'instagram') !== false;

        if ($isInstagram) {
            return Response::create('Can\'t create using Instagram', 400);
        }

        $accessToken = $request->user()->getTwitterToken();
        $accessTokenSecret = $request->user()->getTwitterTokenSecret();
        $twitter->setAccessToken($accessToken, $accessTokenSecret);

        $parsedUrl = parse_url($url);
        $pathParts = explode('/', $parsedUrl['path']);

        $id = array_pop($pathParts);

        try {
            $status = $twitter->getTweetById($id);
            if (empty($status)) {
                return Response::create('Not Found', 404);
            }
            $place = $tranquilPlaceService->createFromTwitterData($id, $status);
            return $place->toArray();
        } catch (\Exception $e) {
            return Response::create('Place exists', 400);
        }
    }

    public function update(Request $request, TranquilPlaceRepository $repository, TranquilPlaceService $tranquilPlaceService)
    {
        $place = $repository->findById($request->route('uid'));
        if (!$place) {
            return Response::create('Not Found', 404);
        }
        $place->setName($request->get('name'));
        $place->setAddress($request->get('address'));
        $place->setPostcode($request->get('postcode'));
        $place->setCaption($request->get('caption'));
        $place->setHidden($request->get('hidden'));
        $place->setEvents($request->get('events'));
        $place->setEventsCaption($request->get('eventsCaption'));
        $place->setThreatened($request->get('threatened'));
        $place->setThreatenedCaption($request->get('threatenedCaption'));
        $place->setVolunteerNeed($request->get('volunteerNeed'));
        $place->setVolunteerNeedCaption($request->get('volunteerNeedCaption'));
        $place->setWalks($request->get('walks'));
        $place->setWalksCaption($request->get('walksCaption'));

        $geolocation = $request->get('geolocation');
        $geolocationP = new Point($geolocation['latitude'], $geolocation['longitude']);

        $place->setGeolocation($geolocationP);
        $pollutionData = $tranquilPlaceService->getPollutionData($place);
        if (!$pollutionData) {
            return Response::create('Couldn\'t find pollution data within 50m of this Tranquil Place.', 404);
        }
        $place->setPollutionData($pollutionData);

        $repository->save($place);

        return $place->toArray();
    }
}
