const angular = require('angular')

const CENTER_LAT = 51.51
const CENTER_LNG = -0.12
const CENTER_ZOOM = 13

const MARKER_SIZE_BY_ZOOM = {
  11: 10,
  12: 15,
  13: 20,
  14: 30,
  15: 35,
  16: 40,
  17: 40
}

angular
  .module('TranquilPavement')
  .component('mapView', {
    template: `
    <navbar on-search="search" on-click-logo="resetCenter"></navbar>
    <main class="main" ng-class="{ 'main--buttons': (view === 'map' && !welcome) }">
      <leaflet defaults="mapConfig" tiles="getSelectedTileset()" maxbounds="bounds" lf-center="center" markers="markers" event-broadcast="events" 
      class="map__container"></leaflet>
      <div ng-switch="view">
        <div ng-switch-when="map">
          <card view="welcome" hide-card="hideCard" ng-if="welcome">
            <img src="/images/logo.png" class="map__welcome-logo">
            <h1 class="map__welcome-title"><strong>Tranquil Pavement London</strong></h1>
            <p>
                An interactive map showcasing tranquil spaces, streets, walks and community hubs in London.
            </p>
            <p>
              Add tranquil spaces by clicking the <strong>+add</strong> button (on mobile only) and 
              posting your photo/video to Instagram or Twitter – include <strong>#tranquilcitylondon</strong> 
              and the location of the space in the description.
            </p>
            <p>
                The map layers use <a href="https://tranquilcity.co.uk/toolkits/">our unique datasets</a> to 
                showcase tranquil and healthier places and streets for all to enjoy. Click the menu option 
                (top right) to select. Click the i button to find out more.
            </p>
            <p>
              Created by <a href="https://tranquilcity.co.uk">Tranquil City</a>, 
              <a href="https://outlandish.com/">Outlandish Cooperative</a> 
              and many beautiful Londoners
            </p>
            <div class="card__divider"></div>
            <div ng-if="locationSupported">
              <div class="card__location-request">
                <img src="/images/icons/my-location.svg">
                <p>Tranquil Pavement would like to use your location</p>
              </div>
              <div class="card__buttons">
                  <button type="button" class="button button--secondary" ng-click="hideCard()">Don't allow</button>
                  <button type="button" class="button button--primary" ng-click="getLocation()">OK</button>
              </div>
            </div>
            <div ng-if="!locationSupported">
                <button type="button" class="button button--primary" ng-click="hideCard()">OK</button>
            </div>
          </card>
        </div>
        <div ng-switch-when="place">
          <place-card place="place" on-close="hideCard" nav="placeNav"></place-card>
        </div>
        <div ng-switch-when="error">
          <card hide-card="hideCard">
            <div class="card__error">
              <p>{{ error }}</p>
              <button type="button" class="button button--primary" ng-click="hideCard()">OK</button>
            </div>
          </card>
        </div>
        <div ng-switch-when="instagram">
          <card hide-card="hideCard">
            <div class="card__instagram">
              <p>
                You can add your tranquil space photos and videos through Instagram or Twitter – just use the 
                <strong>#tranquilcitylondon</strong> tag.
              </p>
              <p>
                <strong>Remember to include your location in the <u>description</u> of the post</strong>, 
                or we won’t be able to place it on the map.
              </p>
              <a id="instagram-link" class="button button--secondary" href="instagram://camera">Instagram</a>
              <a id="twitter-link" class="button button--secondary" href="https://twitter.com">Twitter</a>
            </div>
          </card>
        </div>
        <div ng-switch-when="key">
          <card hide-card="hideCard">
            <div class="card__key">
              <img ng-src="{{ keyFile }}" title="Key">
              <p>
                The map layers help you to discover more tranquil and healthier areas and journeys in your city. 
                They’re based on expert designed indices that are available via our 
                <a href="https://tranquilcity.co.uk/api/">API</a> & <a href="https://tranquilcity.co.uk/toolkits/">GIS Toolkits</a>.
              </p>
              <ul>
                <li>
                  Tranquil City Index (Grid): supports mental restoration, cleaner air and nature connection.
                </li>
                <li>
                  Tranquil City Index (Streets): supports tranquil, calm and mentally restorative journeys.
                </li>
                <li>
                  Healthy Streets Index: supports healthy, active travel & lifestyles.
                </li>
              </ul>
              <p>
                Tranquil Pavement was created in partnership with <a href="https://outlandish.com/">Outlandish Cooperative</a> 
                and everyone who shares to #tranquilcitylondon and were part of our co-creation events. Healthy Streets Index 
                was created by the <a href="https://www.underscorestreets.com/">_Streets</a> collaboration.
              </p>
              <p>
                Phase 1 was funded by the <a href="https://organicity.eu/">OrganiCity</a> project. Phase 2 (2021) was funded by <a href="http://tranquilcity.co.uk/">Tranquil City</a>.
              </p>
              <p><a href="/privacy.html" target="_blank">View our privacy policy.</a></p>
              <div class="key__logos">
                <img class="key__logo" src="/images/logo.png">
                <img class="key__logo" src="/images/organicity.png">
              </div>
            </div>
          </card>
        </div>
      </div>
      <div class="map__buttons">
        <button ng-if="device" class="map__instagram" type="button" ng-click="goToInstagram()">Add +</button>
        <button ng-if="locationSupported" class="map__location" type="button" ng-click="getLocation()"></button>
        <button class="map__info" type="button" ng-click="showKey()"></button>
      </div>
    </main>`,
    controller: ['$scope', '$timeout', '$interval', '$state', '$stateParams', '$transitions', '$window', 'LayersService', 'TranquilPlacesService', 'leafletData',
      function ($scope, $timeout, $interval, $state, $stateParams, $transitions, $window, LayersService, TranquilPlacesService, leafletData) {
        let maxZoom = 17
        let keyFile = '/images/key.jpg'

        LayersService.setTilesets($window.tilesets)
        LayersService.setSelectedTileset($window.tilesets[0])

        angular.extend($scope, {
          view: 'map',
          welcome: false,
          place: null,
          error: false,
          showPollution: true,
          locationSupported: 'geolocation' in navigator,
          keyFile,
          tilesets: $window.tilesets,
          mapConfig: {
            minZoom: 11,
            maxZoom,
            zoomControlPosition: 'bottomright'
          },
          bounds: {
            southWest: {
              lat: 50.0,
              lng: -1.5
            },
            northEast: {
              lat: 51.8,
              lng: 1.5
            }
          },
          center: {
            lat: CENTER_LAT,
            lng: CENTER_LNG,
            zoom: CENTER_ZOOM
          },
          events: {
            map: {
              enable: ['zoomend'],
              logic: 'emit'
            }
          },
          places: [],
          markers: getTubeStations(),
          hideCard: function () {
            window.scrollTo(0, 0)
            angular.element(document.querySelector('.map__container')).css('height', 'auto')
            goToState('app', {}, $state.current.name === 'app')
          },
          search: function (query) {
            return TranquilPlacesService
              .search(query)
              .then(res => {
                setLocation(res.geolocation)
                setPlaces(res.places)
                setCenter(res.geolocation.latitude, res.geolocation.longitude)
                $scope.hideCard()
              }, err => {
                setError(err.data)
              })
          },
          resetCenter: function () {
            setCenter(CENTER_LAT, CENTER_LNG)
            $scope.center.zoom = CENTER_ZOOM
          },
          getLocation: function () {
            console.log('Requesting your location...')
            navigator.geolocation.getCurrentPosition(position => {
              console.log('Your location is (' + position.coords.latitude + ', ' + position.coords.longitude + ')')
              if (!inBounds(position.coords, $scope.bounds)) {
                setError('You are not in London, so we\'re not showing your location.')
                $scope.$apply()
              } else {
                $timeout(() => {
                  $scope.search({latitude: position.coords.latitude, longitude: position.coords.longitude})
                })
                $interval($scope.updateLocation, 15000)
              }
            }, err => {
              console.log('Error', err)
              setError(err)
              $scope.$apply()
            }, { enableHighAccuracy: Boolean($scope.device) })
          },
          getSelectedTileset: LayersService.getSelectedTileset,
          updateLocation: function () {
            navigator.geolocation.getCurrentPosition(position => {
              setLocation({ latitude: position.coords.latitude, longitude: position.coords.longitude })
            })
          },
          goToInstagram: function () {
            updateView('instagram')
            $timeout(() => {
              if (!$scope.appleDevice) {
                document
                  .getElementById('instagram-link')
                  .setAttribute('href', 'intent://instagram.com/_n/mainfeed#Intent;package=com.instagram.android;scheme=https;end')
              }
            }, 100)
          },
          showKey: function () {
            updateView('key')
          }
        })

        /**
         * Event Handlers
         */

        $scope.$on('leafletDirectiveMarker.click', function (event, args) {
          $scope.place = $scope.places.find(place => String(place.uid) === args.modelName)
          if ($scope.place) {
            goToState('app.image', {slug: $scope.place.slug})
          }
        })

        $scope.$on('leafletDirectiveMap.zoomend', function () {
          updateMarkerSize()
        })

        $transitions.onSuccess({}, () => {
          updateView($state.current.name, $stateParams)
        })

        angular.element(document).on('keyup', function (event) {
          if (event.keyCode === 27) {
            $scope.hideCard()
          }
        })

        window.onresize = function () {
          leafletData.getMap().then(map => map.invalidateSize())
        }

        /**
         * Initial setup
         */

        // show welcome message if the user has landed on the default state
        if ($state.current.name === 'app') {
          $scope.welcome = true
        }

        const userAgent = navigator.userAgent.toLowerCase()
        if (/android|webos|iphone|ipad|ipod|blackberry|iemobile|opera mini/i.test(userAgent)) {
          $scope.device = true
          $scope.appleDevice = /iphone|ipad|ipod/i.test(userAgent) && /safari/i.test(userAgent)
        }

        TranquilPlacesService
          .list()
          .then((places) => {
            setPlaces(places)
            updateView($state.current.name, $stateParams)
          })

        /**
         * Functions
         */

        function goToState (state, params, forceUpdate = false) {
          $scope.welcome = false
          $state.go(state, params)
          if (forceUpdate) {
            updateView(state, params)
          }
        }

        function updateView (state, params) {
          $scope.place = null
          $scope.placeNav = null
          if (state === 'app' || state === 'instagram' || state === 'error' || state === 'key') {
            $scope.view = state === 'app' ? 'map' : state
            $timeout(() => {
              const $map = angular.element(document.querySelector('.map__container'))
              const card = angular.element(document.querySelector('.card'))[0]
              if (card) {
                $map.css('height', card.offsetHeight + 'px')
                leafletData.getMap().then(map => map.invalidateSize())
              }
            }, 100)
          } else {
            const placeSlug = params.slug
            let placeIndex = -1
            let checkIndex = 0
            while (placeIndex < 0 && checkIndex < $scope.places.length) {
              const currentPlace = $scope.places[checkIndex]
              if (currentPlace.slug === placeSlug) {
                placeIndex = checkIndex
              }
              checkIndex++
            }
            if (placeIndex > -1) {
              $scope.view = 'place'
              $scope.place = $scope.places[placeIndex]

              const prevPlaceIndex = placeIndex === 0 ? $scope.places.length - 1 : placeIndex - 1
              const nextPlaceIndex = placeIndex === ($scope.places.length - 1) ? 0 : placeIndex + 1
              $scope.placeNav = {
                prev: $scope.places[prevPlaceIndex] && $scope.places[prevPlaceIndex].slug,
                next: $scope.places[nextPlaceIndex] && $scope.places[nextPlaceIndex].slug
              }
            } else {
              $scope.view = 'map'
            }
          }
        }

        function setLocation (geolocation) {
          $scope.markers.me = {
            lat: geolocation.latitude,
            lng: geolocation.longitude,
            message: 'My location',
            focus: false,
            draggable: false,
            icon: {
              iconUrl: '/images/icons/my-location.svg',
              iconSize: [30, 30],
              iconAnchor: [15, 15]
            }
          }
        }

        function setCenter (latitude, longitude) {
          $scope.center.lat = latitude
          $scope.center.lng = longitude
        }

        function setPlaces (places) {
          // remove existing places
          $scope.places.forEach(place => {
            delete $scope.markers[place.uid]
          })

          places = places.filter(p => {
            return !p.hidden &&
              p.geolocation &&
              p.geolocation.latitude &&
              p.geolocation.longitude &&
              inBounds(p.geolocation, $scope.bounds)
          })
          $scope.places = places

          places
            .filter(place => place.geolocation && place.geolocation.latitude && place.geolocation.longitude)
            .forEach(place => {
              $scope.markers[place.uid] = {
                lat: place.geolocation.latitude,
                lng: place.geolocation.longitude,
                focus: false,
                draggable: false,
                icon: {
                  iconUrl: getIcon(place),
                  iconSize: [30, 30],
                  iconAnchor: [15, 30]
                }
              }
            })

          updateMarkerSize()
        }

        function updateMarkerSize () {
          leafletData.getMap().then(map => {
            const zoom = map.getZoom()
            const markerSize = zoom < 11 ? 10 : MARKER_SIZE_BY_ZOOM[zoom]
            Object.keys($scope.markers).forEach(id => {
              const marker = $scope.markers[id]
              if (id.indexOf('tube_') === 0) {
                marker.icon.iconSize = zoom < 14 ? [0, 0] : [markerSize / 3, markerSize / 3]
                marker.icon.iconAnchor = [markerSize / 6, markerSize / 6]
              } else {
                marker.icon.iconSize = [markerSize, markerSize]
                marker.icon.iconAnchor = [markerSize / 2, markerSize]
              }
            })
          })
        }

        function getTubeStations () {
          const markers = {}
          window.TUBE_STATIONS.features.forEach(station => {
            const name = station.properties.name.replace(/-/g, ' ')
            markers['tube_' + name] = {
              lat: station.geometry.coordinates[1],
              lng: station.geometry.coordinates[0],
              focus: false,
              draggable: false,
              icon: {
                iconUrl: '/images/icons/tube-station.svg',
                iconSize: [30, 30],
                iconAnchor: [15, 15]

              },
              title: name,
              message: name
            }
          })
          return markers
        }

        function getIcon (place) {
          if (place.events) {
            return '/images/icons/map-pin-events.svg'
          }

          if (place.threatened) {
            return '/images/icons/map-pin-threatened.svg'
          }

          if (place.volunteerNeed) {
            return '/images/icons/map-pin-volunteer-need.svg'
          }

          if (place.walks) {
            return '/images/icons/map-pin-walks.svg'
          }

          return '/images/icons/map-pin.svg'
        }

        function inBounds (coords, bounds) {
          return coords.latitude > bounds.southWest.lat &&
            coords.latitude < bounds.northEast.lat &&
            coords.longitude < bounds.northEast.lng &&
            coords.longitude > bounds.southWest.lng
        }

        function setError (error) {
          $scope.error = (typeof error === 'string') ? error : error.message
          updateView('error')
        }
      }]
  })
