<?php
/**
 * Created by PhpStorm.
 * User: Joaquim
 * Date: 17/12/2017
 * Time: 14:20
 */

namespace TranquilCity\TranquilPavement\Application\Services;


use TranquilCity\TranquilPavement\Domain\Models\Point;
use TranquilCity\TranquilPavement\Domain\Models\TranquilPlace;
use TranquilCity\TranquilPavement\Domain\Repositories\PollutionDataPointRepository;
use TranquilCity\TranquilPavement\Domain\Repositories\TranquilPlaceRepository;

class TranquilPlaceService
{

    private $pollution;
    private $tranquilPlaces;

    public function __construct(TranquilPlaceRepository $tranquilPlaces,
                                PollutionDataPointRepository $pollution,
                                Geocoder $geocoder)
    {
        $this->tranquilPlaces = $tranquilPlaces;
        $this->pollution = $pollution;
        $this->geocoder = $geocoder;
        $this->places = array_map(
            function ($arr) {
                return $arr[0];
            },
            array_map('str_getcsv', file(__DIR__ . '/../../../data/locations.csv'))
        );
    }

    public function createFromInstagramData($data)
    {
        $slug = $data->slug;

        $place = new TranquilPlace();
        $place->setExternalUrl($data->permalink);
        $place->setName('Unknown place');
        $place->setSlug($slug);
        $place->setInstagramId($data->id);
        $place->setImageUrl($data->media_url);
        if (property_exists($data, 'like_count')) {
            $place->setNumLikes($data->like_count);
        }
        if (property_exists($data, 'comments_count')) {
            $place->setNumComments($data->comments_count);
        }
        if (property_exists($data, 'caption') && $data->caption) {
            $text = $data->caption;
            $place->setCaption($text);
            $name = $place->getCaptionName();
            if ($name) {
                $place->setName($name);
            }
            $this->geocode($place);
        }

        $place->setCreatedTime(time());
        $pollutionData = $this->getPollutionData($place);
        if ($pollutionData) {
            $place->setPollutionData($pollutionData);
        }
        $this->tranquilPlaces->add($place);
        return $place;
    }

    /**
     * @param $twitterId
     * @param $data
     * @return TranquilPlace
     * @throws \Exception
     */
    public function createFromTwitterData($twitterId, $data)
    {
        if (property_exists($data, 'retweeted_status') && $data->retweeted_status) {
            throw new \Exception('This is a retweet, skipping...');
        }
        $author = $data->user->screen_name;
        $authorImageUrl = $data->user->profile_image_url_https;
        $slug = "/${author}/status/${twitterId}";

        $place = new TranquilPlace();
        $place->setName('Unknown place');

        $place->setExternalUrl("https://twitter.com${slug}");

        $caption = '';
        if (property_exists($data, 'full_text')) {
            $caption = $data->full_text;
        } else if (property_exists($data, 'text')) {
            $caption = $data->text;
        }
        $caption = preg_replace('%https://t.co/[^ ]+%', '', $caption);
        $place->setCaption($caption);

        $this->updatePlaceWithVariableTwitterData($place, $data);

        if (!$place->hasGeolocation()) {
            $this->geocode($place);
        }

        $place->setSlug($twitterId);
        $place->setTwitterId($twitterId);
        $place->setAuthor($author);
        $place->setAuthorImageUrl($authorImageUrl);
        $place->setCreatedTime(strtotime($data->created_at));

        $this->tranquilPlaces->add($place);
        return $place;
    }


    /**
     * @param TranquilPlace $place
     * @param $data
     * @return TranquilPlace
     * @throws \Exception
     */
    public function updateFromTwitterData(TranquilPlace $place, $data)
    {
        $this->updatePlaceWithVariableTwitterData($place, $data);
        $this->tranquilPlaces->save($place);
        return $place;
    }

    /**
     * @param TranquilPlace $place
     * @param $data
     * @throws \Exception
     */
    private function updatePlaceWithVariableTwitterData(TranquilPlace $place, $data)
    {
        if (!property_exists($data->entities, 'media') || !$data->entities->media) {
            throw new \Exception('Warning: no media for place ' . $place->getExternalUrl());
        }

        $media = $data->entities->media;

        $place->setImageUrl($media[0]->media_url_https);
        $place->setAuthorImageUrl($data->user->profile_image_url_https);

        $place->setNumLikes($data->favorite_count);
        $place->setNumRetweets($data->retweet_count);

        $name = $place->getName();
        $captionName = $place->getCaptionName();
        if ($name === 'Unknown place' && $captionName) {
            $place->setName($captionName);
        }

        if (!$place->hasGeolocation()) {
            if (property_exists($data, 'place') && $data->place) {
                $place->setName($data->place->name);
                $place->setGeolocation(Point::fromBoundingBox($data->place->bounding_box->coordinates));
            }

            if (property_exists($data, 'coordinates') && $data->coordinates) {
                $place->setGeolocation(Point::fromArray(array_reverse($data->coordinates->coordinates)));
            }
        }

        $pollutionData = $this->getPollutionData($place);
        if ($pollutionData) {
            $place->setPollutionData($pollutionData);
        }
    }

    public function getPollutionData(TranquilPlace $place)
    {
        $location = $place->getGeolocation();
        if ($location && $location->getLatitude() && $location->getLongitude()) {
            $pollutionData = $this->pollution->findByLocation($location);
            if ($pollutionData) {
                return $pollutionData;
            }
        }
        return null;
    }

    public function geocode(TranquilPlace $place)
    {
        $captionLines = $place->getCaptionLines();

        $fullAddress = implode(', ', $captionLines);

        $postcodeRegex = '#[A-Z]{1,2}[0-9][0-9A-Z]?\s?[0-9][A-Z]{2}#i';
        $streetRegex = '#\w+ ' .
            '(' .
            'Road|Rd|Street|St|Way|Avenue|Ave|Drive|Dr|Grove|Lane|Gardens|Place|Circus|Crescent|' .
            'Close|Square|Hill|Mews|Vale|Rise|Row|Mead|Wharf|End|Court|Cross|Side|View|Walk|Park|Meadow|' .
            'Pond|Lake' .
            ')' .
            '\b#i';
        $riverRegex = '#River \w+\b#i';
        $addressLines = [];
        $postcode = '';
        foreach ($captionLines as $line) {
            if (strpos($line, '#') !== false) {
                continue;
            }
            $shortMatch = false;
            if (strlen($line) > 30) {
                $shortMatch = true;
            }
            preg_match_all($streetRegex, $line, $matches);
            if ($matches && count($matches) && count($matches[0])) {
                $addressLines[] = $shortMatch ? implode($matches[0], ', ') : $line;
            }
            preg_match($riverRegex, $line, $matches);
            if ($matches && count($matches)) {
                $addressLines[] = $matches[0];
            }
            preg_match($postcodeRegex, $line, $matches);
            if ($matches && count($matches)) {
                $postcode = $matches[0];
            }
        }

        $area = '';
        foreach ($this->places as $p) {
            preg_match('#' . $p . '#i', $fullAddress, $matches);
            if ($matches && count($matches)) {
                $area = $p;
            }
        }
        if ($area) {
            $addressLines[] = $area;
        }
        if ($postcode) {
            $addressLines[] = $postcode;
        }
        $address = implode($addressLines, ', ');
        if (!$address) {
            return;
        }

        $coords = $this->geocoder->geocode($address, false);
        if (!$coords) {
            error_log('Geocoding failed: ' . $address);
            return;
        }

        $place->setGeolocation($coords);
        if (count($addressLines) < 2) {
            error_log('Geocoding poor match: ' . $address);
            $place->setHidden(true);
        }
    }

    /**
     * @param TranquilPlace[] $places
     */
    public function order($places)
    {
        foreach ($places as $place) {
            if ($place->hasGeolocation()) {
                $center = $place->getGeolocation();
                $nearby = $this->tranquilPlaces->findByLocation($center);
                if (count($nearby)) {
                    $defaultIndex = null;
                    foreach ($nearby as $placeAndDistance) {
                        /** @var TranquilPlace $nearest */
                        $nearest = $placeAndDistance['place'];
                        $defaultIndex = $nearest->getDefaultIndex();
                        if ($defaultIndex) {
                            break;
                        }
                    }
                    $place->setDefaultIndex($defaultIndex);
                    $this->tranquilPlaces->save($place);
                }
            }
        }
    }

    public function reorder()
    {
        $places = $this->tranquilPlaces->findAllLocated();
        $next = array_pop($places);
        /** @var TranquilPlace[] $newOrder */
        $newOrder = [ $next ];
        while (count($places)) {
            $next = $this->removeNearest($next, $places);
            $newOrder[] = $next;
        }
        foreach ($newOrder as $i => $place) {
            $place->setDefaultIndex($i);
            $this->tranquilPlaces->save($place);
        }
    }

    /**
     * @param TranquilPlace $place
     * @param TranquilPlace[] $places
     * @return TranquilPlace $place
     */
    private function removeNearest(TranquilPlace $place, &$places)
    {
        $center = $place->getGeolocation();
        $minDistance = null;
        $index = null;
        foreach ($places as $i => $p) {
            $compare = $p->getGeolocation();
            $distance = $this->getDistance($center, $compare);
            if ($minDistance === null || $distance < $minDistance) {
                $minDistance = $distance;
                $index = $i;
            }
        }
        $nearest = array_splice($places, $index, 1)[0];
        return $nearest;
    }

    private function getDistance(Point $from, Point $to)
    {
        $latitudeFrom = $from->getLatitude();
        $latitudeTo = $to->getLatitude();
        $longitudeFrom = $from->getLongitude();
        $longitudeTo = $to->getLongitude();
        // convert from degrees to radians
        $latFrom = deg2rad($latitudeFrom);
        $lonFrom = deg2rad($longitudeFrom);
        $latTo = deg2rad($latitudeTo);
        $lonTo = deg2rad($longitudeTo);

        $latDelta = $latTo - $latFrom;
        $lonDelta = $lonTo - $lonFrom;

        $angle = 2 * asin(sqrt(pow(sin($latDelta / 2), 2) +
                cos($latFrom) * cos($latTo) * pow(sin($lonDelta / 2), 2)));
        return $angle * 6371000;
    }
}
