<?php

namespace TranquilCity\TranquilPavement\Infrastructure\Doctrine\Repositories;

use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Query\ResultSetMappingBuilder;
use TranquilCity\TranquilPavement\Domain\Models\Point;
use TranquilCity\TranquilPavement\Domain\Models\TranquilPlace;
use TranquilCity\TranquilPavement\Domain\Repositories\TranquilPlaceRepository;

class DoctrineTranquilPlaceRepository implements TranquilPlaceRepository
{

    private $entityManager;
    private $repository;

    /**
     * Constructs the DoctrineTranquilPlaceRepository
     *
     * @param EntityManager $entityManager The Doctrine Entity Manager (autowired)
     *
     */
    public function __construct(EntityManager $entityManager)
    {
        $this->entityManager = $entityManager;
        $this->repository = $this->entityManager->getRepository(TranquilPlace::class);
    }

    public function add(TranquilPlace $place)
    {
        $this->entityManager->persist($place);
        $this->entityManager->flush();
    }

    public function save(TranquilPlace $place)
    {
        $this->entityManager->persist($place);
        $this->entityManager->flush();
    }

    public function remove(TranquilPlace $place)
    {
        $this->entityManager->remove($place);
        $this->entityManager->flush();
    }

    public function findAll($order = 'defaultIndex')
    {
        return $this->repository->findBy([], [ $order => 'DESC' ]);
    }

    public function findAllLocated()
    {
        return $this->repository->createQueryBuilder('p')
            ->where('p.geolocation IS NOT NULL')
            ->getQuery()
            ->getResult();
    }

    public function findAllNotLocated()
    {
        return $this->repository->createQueryBuilder('p')
            ->where('p.geolocation IS NULL')
            ->getQuery()
            ->getResult();
    }

    public function findAllInstagram()
    {
        return $this->repository->createQueryBuilder('p')
            ->where('p.instagramId IS NOT NULL')
            ->getQuery()
            ->getResult();
    }

    public function findAllTwitter()
    {
        return $this->repository->createQueryBuilder('p')
            ->where('p.twitterId IS NOT NULL')
            ->getQuery()
            ->getResult();
    }

    public function findById($uid)
    {
        return $this->repository->find($uid);
    }

    public function findByThirdPartyId($id)
    {
        $results = $this->repository->createQueryBuilder('p')
            ->where("p.instagramId = '${id}'")
            ->orWhere("p.slug = '${id}'")
            ->orWhere("p.twitterId = '${id}'")
            ->getQuery()
            ->getResult();
        return count($results) ? $results[0] : null;
    }

    public function findByLocation(Point $geolocation, $order = 'distance')
    {
        $rsm = new ResultSetMappingBuilder($this->entityManager);
        $rsm->addRootEntityFromClassMetadata(TranquilPlace::class, 'p0_');
        $rsm->addScalarResult('distance', 'distance', 'float');

        if ($order !== 'distance') {
            $order = 'p0_.' . $this->entityManager->getClassMetadata(TranquilPlace::class)->getColumnName($order);
        }

        $sql = "
        SELECT p0_.*, ST_AsText(p0_.geolocation) AS geolocation, ST_DistanceSphere(p0_.geolocation, ST_Point(:x, :y)) AS distance
        FROM places as p0_ WHERE hidden IS FALSE AND geolocation IS NOT NULL
        ORDER BY $order
        ";

        $parameters = [
            'x' => (string)$geolocation->getLongitude(),
            'y' => (string)$geolocation->getLatitude()
        ];

        $query = $this->entityManager->createNativeQuery(
            $sql,
            $rsm
        );

        $query->setParameters($parameters);

        $result = $query->getResult();

        return array_map(function ($row) {
            return ['place' => $row[0], 'distance' => $row['distance']];
        }, $result);
    }

    /**
     * Not used, kept in case we want to go back to searching within a radius (instead of finding the N closest)
     */
    public function findWithinRadius(Point $geolocation)
    {
        $radius = 3;

        $rsm = new ResultSetMappingBuilder($this->entityManager);
        $rsm->addRootEntityFromClassMetadata(TranquilPlace::class, 'p0_');
        $rsm->addScalarResult('distance', 'distance', 'float');

        $sql = "SELECT 
                  *, 
                  ST_AsText(p0_.geolocation) AS geolocation,
                  ST_DistanceSphere(ST_Point(:x, :y), p0_.geolocation) AS distance
                FROM places p0_ WHERE
                  ST_Contains(
                    ST_MakeEnvelope(
                      :x - :radius / (69 * COS(RADIANS(:y))), cast(:y AS DOUBLE PRECISION) - cast(:radius AS DOUBLE PRECISION) / 69,
                      :x + :radius / (69 * COS(RADIANS(:y))), cast(:y AS DOUBLE PRECISION) + cast(:radius AS DOUBLE PRECISION) / 69,
                      4326
                    ),
                    p0_.geolocation
                  )
                  AND ST_DistanceSphere(ST_Point(:x, :y), p0_.geolocation) <=  cast(:radius AS DOUBLE PRECISION) * 1000";

        $radiusKm = $radius * 1.60934;
        $parameters = [
            'x' => (string)$geolocation->getLongitude(),
            'y' => (string)$geolocation->getLatitude(),
            'radius' => (string)$radiusKm
        ];

        $query = $this->entityManager->createNativeQuery(
            $sql,
            $rsm
        );

        $query->setParameters($parameters);

        $result = $query->getResult();

        return $result;
    }

}