<?php

namespace TranquilCity\TranquilPavement\Infrastructure\Services;

use Facebook\Exceptions\FacebookSDKException;
use Facebook\Facebook;

Class TranquilFacebook
{

    private $fb;
    private $instagramBusinessId;
    private $accessToken;

    public function __construct($args)
    {
        $FIELDS = 'id,media_type,comments_count,like_count,caption,media_url,permalink';
        $this->fields = $FIELDS . ',children{media_url,permalink}';
        try {
            $this->fb = new Facebook([
                'app_id' => $args['apiId'],
                'app_secret' => $args['apiSecret'],
                'default_graph_version' => 'v3.2'
            ]);
            $this->accessToken = $args['apiKey'];
            $this->instagramBusinessId = $args['instagramBusinessId'];
        } catch (FacebookSDKException $e) {
            error_log($e->getMessage());
        }
    }

    public function getBusinesses($accessToken)
    {
        $businesses = [];
        try {
            $response = $this->fb->sendRequest(
                'get',
                '/me',
                [
                    'fields' => 'accounts',
                ],
                $accessToken
            );
            $body = $response->getDecodedBody();
            $pageIds = [];
            if (!empty($body['accounts']['data'])) {
                foreach ($body['accounts']['data'] as $page) {
                    $pageIds[] = $page['id'];
                }
            }
            $businessIds = [];
            foreach ($pageIds as $pageId) {
                $response = $this->fb->sendRequest(
                    'get',
                    '/' . $pageId,
                    [
                        'fields' => 'instagram_business_account',
                    ],
                    $accessToken
                );
                $body = $response->getDecodedBody();
                if (!empty($body['instagram_business_account'])) {
                    $businessIds[] = $body['instagram_business_account']['id'];
                }
            }
            foreach ($businessIds as $businessId) {
                $response = $this->fb->sendRequest(
                    'get',
                    '/' . $businessId,
                    [
                        'fields' => 'profile_picture_url,name',
                    ],
                    $accessToken
                );
                $body = $response->getDecodedBody();
                if (!empty($body)) {
                    $businesses[] = $body;
                }
            }

        } catch (FacebookSDKException $e) {
            error_log($e->getMessage());
        }
        return $businesses;
    }

    /**
     * Get top tagged media
     *
     * @param string $name Valid tag name
     * @param $businessId
     * @param $accessToken
     * @return mixed
     */
    public function getTagMedia($name, $businessId = null, $accessToken = null)
    {
        $media = [];
        $businessId = $businessId ?: $this->instagramBusinessId;
        $accessToken = $accessToken ?: $this->accessToken;
        try {
            $response = $this->fb->sendRequest(
                'get',
                '/ig_hashtag_search',
                [
                    'user_id' => $businessId,
                    'q' => $name
                ],
                $accessToken
            );
            $body = $response->getDecodedBody();
            $tagId = null;
            if (!empty($body['data'][0]['id'])) {
                $tagId = $response->getDecodedBody()['data'][0]['id'];
            }
            if (!$tagId) {
                error_log('Tag not found');
                return $media;
            }

            $top = $this->getTagMediaTop($tagId, $businessId, $accessToken);
            $media = array_merge($media, $top->asArray());

            $next = $this->getTagMediaPaginated($tagId, $businessId, $accessToken);
            $lastAfter = null;
            while (true) {
                $metaData = $next->getMetaData();

                if (!empty($metaData['paging']['cursors']['after'])) {
                    $after = $metaData['paging']['cursors']['after'];
                    if ($after && $after === $lastAfter) {
                        break;
                    }
                    $lastAfter = $after;
                }

                $media = array_merge($media, $next->asArray());


                if (empty($metaData['paging']['next'])) {
                    break;
                }

                $next = $this->getTagMediaPaginated($tagId, $businessId, $accessToken, $after);
            }
        } catch (FacebookSDKException $e) {
            error_log($e->getMessage());
        }

        $images = [];
        foreach ($media as $m) {
            if ($m['media_type'] === 'IMAGE') {
                $images[] = $m;
            } else if ($m['media_type'] === 'CAROUSEL_ALBUM') {
                foreach ($m['children'] as $i => $child) {
                    $copy = $m;
                    $copy['media_url'] = $child['media_url'];
                    $copy['id'] = $child['id'];
                    $copy['permalink'] = $child['permalink'];
                    $images[] = $copy;
                }
            }
        }
        foreach ($images as &$image) {
            $link = $image['permalink'];
            preg_match('#p/(.+)/#', $link, $matches);
            $image['slug'] = $matches[1];
        }

        return $images;
    }

    /**
     * @param $tagId
     * @param $businessId
     * @param $accessToken
     * @return \Facebook\GraphNodes\GraphEdge
     * @throws FacebookSDKException
     */
    private function getTagMediaTop ($tagId, $businessId, $accessToken) {
        $response = $this->fb->sendRequest(
            'get',
            '/' . $tagId . '/top_media',
            [
                'user_id' => $businessId,
                'fields' => $this->fields
            ],
            $accessToken
        );
        return $response->getGraphEdge();
    }

    /**
     * @param $tagId
     * @param $businessId
     * @param $accessToken
     * @param $after
     * @return \Facebook\GraphNodes\GraphEdge
     * @throws FacebookSDKException
     */
    private function getTagMediaPaginated ($tagId, $businessId, $accessToken, $after = null) {
        $response = $this->fb->sendRequest(
            'get',
            '/' . $tagId . '/recent_media',
            [
                'user_id' => $businessId,
                'fields' => $this->fields,
                'after' => $after
            ],
            $accessToken
        );
        return $response->getGraphEdge();
    }

    /**
     * @param $id
     * @param $accessToken
     * @return array
     * @throws FacebookSDKException
     */
    private function getMediaById ($id, $accessToken) {
        $response = $this->fb->sendRequest(
            'get',
            '/' . $id,
            ['fields' => $this->fields],
            $accessToken
        );
        return $response->getGraphNode()->asArray();
    }
}